/* Fast-Fourier-Transform in Fork95.   C.W.Kessler 08/96 */

#include <fork.h>
#include <io.h>
#include <math.h>
#include <stdlib.h>

extern void shfree( void * );

typedef struct {
  double re, im;
} cplx_t, *cplx;

/* The following routines for complex arithmetics
 * will later be available in math.h as assembler routines:
 */

async void cinit(cplx rop, pr double re, pr double im)
{
  rop->re = re;
  rop->im = im;
}

async void ccpy(cplx rop, pr cplx op)
{
  rop->re = op->re;
  rop->im = op->im;
}

async void cadd(cplx rop, pr cplx op1, pr cplx op2)
{
  rop->re = op1->re + op2->re;
  rop->im = op1->im + op2->im;
}

async void csub(cplx rop, pr cplx op1, pr cplx op2)
{
  rop->re = op1->re - op2->re;
  rop->im = op1->im - op2->im;
}

async void cmul(cplx rop, pr cplx op1, pr cplx op2)
{
  pr cplx_t t;

  t.re = op1->re * op2->re - op1->im * op2->im;
  t.im = op1->im * op2->re + op1->re * op2->im;

  rop->re = t.re;
  rop->im = t.im;
}

sync void seq_fft(sh cplx out, sh cplx in, sh int n, sh cplx w)
{
  /* seq. recursive FFT implementation for complex array in[].
   * Assume n is a power of 2. Operand array in is not corrupted.
   */
 sh cplx even, odd, fteven, ftodd;   /*temporary pointers*/
 sh int ndiv2;
 pr int i;

 if (n == 1)  {
    seq ccpy(out, in);
    return;
 }

 /* the general case: compute temporary arrays even[], odd[]
  *                   and in parallel call fft for each of them
  */
 ndiv2   = n >> 1;
 even    = (cplx)shalloc(ndiv2 * sizeof(cplx_t));  /* there is more space on the */
 odd     = (cplx)shalloc(ndiv2 * sizeof(cplx_t));  /* shared heap than on private */
 fteven  = (cplx)shalloc(ndiv2 * sizeof(cplx_t));  /* there is more space on the */
 ftodd   = (cplx)shalloc(ndiv2 * sizeof(cplx_t));  /* shared heap than on private */
 for(i = 0; i < ndiv2; i++) {
   even[i].re = in[2 * i].re;
   even[i].im = in[2 * i].im;
   odd[i].re  = in[2 * i + 1].re;
   odd[i].im  = in[2 * i + 1].im;
 }
 seq_fft(fteven, even, ndiv2, w);
 seq_fft(ftodd,  odd,  ndiv2, w);

 farm 
  for(i = 0; i < ndiv2; i++ ) {   /*sequential loop*/
    pr cplx_t t;
    cmul(&t, &w[i], &ftodd[i]);
    cadd(&out[i],         &fteven[i], &t);
    csub(&out[i + ndiv2], &fteven[i], &t);
  }
 shallfree();   /* free even, odd */
 return;
}

sync void fft(sh cplx out, sh cplx in, sh int n, sh cplx w)
{
  /* recursive FFT implementation for complex array in[].
   * complex array w[] contains the powers of the primitive
   * n'th root w of unity: { 1, w, w^2, w^3, ..., w^(n-1) }.
   * Assume n is a power of 2.
   * Operand array in is not corrupted.
   */
 sh cplx even, odd, fteven, ftodd;   /*temporary pointers*/
 sh int p = 0;
 sh int ndiv2;
 pr int i;

 $ = mpadd( &p, 1 );  /* ensure consecutive numbering of procs */

 seq  prS("fft\n");   /*diagnosis output*/

 if (n == 1) {
    seq ccpy(out, in);
    return;
 }
 if (p == 1) {
    seq_fft(out, in, n, w);
    return;
 }

 /* general case: compute temporary arrays even[], odd[]    *
  *               and in parallel call fft for each of them */
 ndiv2  = n >> 1;
 even   = (cplx)shalloc(ndiv2 * sizeof(cplx_t));
 odd    = (cplx)shalloc(ndiv2 * sizeof(cplx_t));
 fteven = (cplx)shalloc(ndiv2 * sizeof(cplx_t));
 ftodd  = (cplx)shalloc(ndiv2 * sizeof(cplx_t));
 for(i = $; i < ndiv2; i += p) {     /*dataparallel loop*/
   even[i].re = in[2 * i].re;
   even[i].im = in[2 * i].im;
   odd[i].re  = in[2 * i + 1].re;
   odd[i].im  = in[2 * i + 1].im;
 }
 if ($ < p / 2)  /*split current group into two equally sized subgroups:*/
      fft(fteven, even, ndiv2, w);
 else
      fft(ftodd,  odd,  ndiv2, w);

 farm
  for(i = $; i < ndiv2; i += p) {     /*dataparallel loop*/
    pr cplx_t t;
    cmul(&t, &w[i], &ftodd[i]);
    cadd(&out[i],         &fteven[i], &t);
    csub(&out[i + ndiv2], &fteven[i], &t);
  }
 shallfree();  /* free even, odd */
 return;
}

sync void FFT(sh cplx out, sh cplx in, sh int n, sh cplx w)
{
 sh cplx pow = (cplx)shalloc(n * sizeof(cplx_t));
 sh cplx_t wtoi, newwtoi;
 sh int p = groupsize();
 sh int twotoi, j;

 /* compute vector pow[] of powers of w: { 1, w, w^2, ..., w^(n-1) } */
 seq {
    ccpy(&wtoi, w);
    cinit(&pow[0], 1.0, 0.0);
    ccpy(&pow[1], w);
 }

 for (twotoi = 2; twotoi < n; twotoi *= 2) {
   seq cmul(&wtoi, &wtoi, &wtoi);
   farm
     forall (j, 0, twotoi, p)
       cmul(&pow[twotoi + j], &wtoi, &pow[j]);
 }

 fft(out, in, n, pow);   /* call recursive fft routine */

 shallfree();
}

int
main(int ac, char *av[])
{
 start {
  sh int n = 16;
  sh cplx a, ft;
  sh cplx_t w;
  sh unsigned int starttime, stoptime;
  sh double t;
  pr int i;

  seq {

     if (ac > 1) {
        n = atoi(av[1]);
     }

     t = 2 * PI / (double)n;

     a  = (cplx)shmalloc(n * sizeof(cplx_t));
     ft = (cplx)shmalloc(n * sizeof(cplx_t));
     cinit(&w, cos(t), sin(t));

#if 0
     for (i = 0; i < n; i++)
        cinit(&a[i], cos((double)i), sin((double)i));
#endif
      cinit(&a[ 0], -5,     0);
      cinit(&a[ 1], -1.55,  0);
      cinit(&a[ 2], -0.939, 0);
      cinit(&a[ 3],  1.6,   0);
      cinit(&a[ 4],  3,     0);
      cinit(&a[ 5],  3.51,  0);
      cinit(&a[ 6],  3.77,  0);
      cinit(&a[ 7],  1.66,  0);
      cinit(&a[ 8], -1.5,   0);
      cinit(&a[ 9], -2.7,   0);
      cinit(&a[10], -3.06,  0);
      cinit(&a[11], -3.02,  0);
      cinit(&a[12], -1.00,  0);
      cinit(&a[13],  0.736, 0);
      cinit(&a[14],  0.232, 0);
      cinit(&a[15], -0.25,  0);

     for (i = 0; i < n; i++)
        printf("%5.2f%+5.2fi\n", a[i].re, a[i].im);

     printf("Compute FFT:\n");
  }

  farm starttime = getct();
  FFT(ft, a, n, &w);
  farm stoptime = getct();
  seq {
     for (i = 0; i < n; i++) 
        printf("%5.2f%+5.2fi\n", ft[i].re, ft[i].im);
  }
  seq printf("Time: %u PRAM CPU cycles\n", stoptime - starttime);

 }
 barrier;
 return 0;
}
