/** Gauss-Elimination in Fork.
 *  C.W.Kessler, reworked 12/99
 *  Uses Matrix data structure, reduce skeleton function, 
 *  forall2 loop macro.
 *  To be linked with /util/Matrix.o and /util/skeletons.o.
 */

#include <fork.h>
#include <io.h>
#include <math.h>
#include <assert.h>
#include <Matrix.h>
#include <skeletons.h>

/* The forall() macro cannot be nested.
 * We apply explicit flattening to a self-scheduling parallel loop
 * to exploit 2D parallelism.
 */

#define GLOBALHEAP 0
#define GROUPHEAP 1

sync Matrix build_system( sh int *pn ) 
{
   /* build initial system of *pn equations in *pm variables.
    * RHS vector stored in column *pn of system matrix.
    */
  sh int n;
  sh Matrix s;
  int i;
  n = 3;    /*example values*/
  s = new_Matrix( n, n+1, GLOBALHEAP );
  Set(s,0,0, 2.0);   Set(s,0,1, -1.0);  Set(s,0,2, 3.0);  Set(s,0,3, 6.0);
  Set(s,1,0, -3.0);  Set(s,1,1, 3.0);   Set(s,1,2, -2.0); Set(s,1,3, 4.0);
  Set(s,2,0, 1.0);   Set(s,2,1, 2.0);   Set(s,2,2, 1.0);  Set(s,2,3, 0.0);
  *pn = n;
  return s;
}


void print_system( Matrix s )
{
  /* One processor prints the system of n equations
   * and m variables in sequential. */
  int i, j;
  printf("New %d x %d system:\n", s->nrows, s->ncols );
  printf("==================================================\n");
  for (i=0; i<s->nrows; i++) {
     printf("EQ%d: ", i);
     for (j=0; j<s->ncols-1; j++) 
        printf(" %5.2lf x%.2d", Get(s,i,j), j );
     printf(" == %5.2lf\n", Get(s,i,s->ncols-1) );
  }
  printf("==================================================\n");
}


typedef struct { float val; int loc; } pair;

straight void fabsmax( void *m, void *x, void *y )
{
 farm
  if (fabs(((pair*)x)->val) >= fabs(((pair*)y)->val)) {
    ((pair*)m)->val = ((pair *)x)->val;  ((pair*)m)->loc = ((pair*)x)->loc;
  }
  else {
    ((pair*)m)->val = ((pair *)y)->val;  ((pair*)m)->loc = ((pair*)y)->loc;
  }
}


sync void pivotize( sh Matrix a, sh int n, sh int r )
{
  sh float *hlp;
  sh pair piv;
  sh pair *col = (pair *)shalloc((n-r)*sizeof(pair));
  int i;
  farm
   forall (i, 0, n-r, #) {
     col[i].loc = r+i;
     col[i].val = Get(a,i+r,r);
   }
  reduce( fabsmax, &piv, (void **)col, n-r, sizeof(pair) );

  seq printf("Pivot row is %d, pivot is %f\n", piv.loc, piv.val);

  // permute rows:
  hlp = a->row[r];
  a->row[r] = a->row[piv.loc];
  a->row[piv.loc] = hlp;
}


sync void eliminate( sh Matrix a, sh int n )
{
  /* applies Gaussian elimination to system a ([n]X[n+1]).
   * Assume that matrix a is regular.
   * For a system of n equations in n variables,
   * up to n^2 processors could be used to update in parallel.
   * The solution is returned in the right hand side column a[:][n] 
   * of the modified system tableau.
   */
 int i, j, ij;
 sh int r;
 sh int p = #;
 sh float factor;
 sh float *rowfactor = (float *)shalloc( n * sizeof(float));
 
 for (r=0; r<n; r++) {  /* do n elimination steps */ 

    if (r<n-1)  pivotize( a, n, r );

    seq print_system( a );

    factor = 1.0 / Get(a,r,r);

    seq printf("Stage %d factor %lf p=%d\n", r, factor, p);

    /* normalize equation r: */
    farm
      forall ( j, r, n+1, p )
        Set(a,r,j, Get(a,r,j) * factor);

    seq print_system( a );

    farm
      forall ( i, 0, n, p ) 
        rowfactor[i] = Get(a,i,r);   // copying avoids data dependencies

    seq
      for (j=r; j<n; j++) 
        printf("->rowfactor[%d]=%f\n", j, rowfactor[j]);

    farm {
     forall2( i, j, ij, 0, n, r, n+1, p ) {
      //i = ij / (n-r+1);  // row index of element to be updated
      if (i!=r) {
        pprintf("update (%d,%d)\n", i, j);
        /* update row i: */
        //j = r + ij % (n-r+1);  // column index of element to be updated
        Set( a,i,j, Get(a,i,j) - Get(a,r,j) * rowfactor[i] ); 
      }
     }
    }
    seq print_system( a );
 }  /* for r */
 return;
}


sh Matrix s;  /* system tableau */
sh int n;     /* n eq's in m var's */

void main( void )
{
  pr int starttime, stoptime;
  start {
    s = build_system( &n );
    seq print_system( s );
    farm starttime = getct();
    eliminate( s, n );
  }
  stoptime = getct();
  if ($==0) printf("\nTime: %d PRAM CPU cycles\n", stoptime - starttime);
  barrier;
}
