/* Gauss-Elimination in Fork95.   C.W.Kessler 02/96 */

#include <fork.h>
#include <io.h>
#include <math.h>
#include <assert.h>
#include <string.h>     // prototype for memcpy() - important!

/* The forall() macro cannot be nested.
 * We apply explicit flattening to a self-scheduling parallel loop
 * to exploit 2D parallelism.
 */


sync float **build_system( sh int N )
{
   /* build initial system of *pn equations in *pm variables.
    * RHS vector stored in column *pn of system matrix.
    */
  pr int i, j;
  pr int n;
  sh float **s;
  sh int p = groupsize();
  n = N;
  seq s = (float **) shmalloc( n * sizeof( float *));

  farm {
    forall (i, 0, n, p) {
      s[i] = (float *) shmalloc( n * sizeof( float ));
      for (j = 0; j < n; j++)
        s[i][j] = (j >= n - i) ? 1 : i + 1;
    }
  }

  return s;
}

#ifdef HAVE_IO
async void print_system(float **M, int N)
{
  /* One processor prints the system of n equations
   * and m variables in sequential. */
  pr int i, j;
  for (i = 0; i < N; i++) {
     for (j = 0; j < N; j++) 
        printf(" %12.6f", M[i][j]);
     printf("\n");
  }
}
#endif


typedef struct { float val; int loc; } pair;

sync void reduce( async sh void (*f)(void *, void *, void *),
                  sh void *s, sh void **d, sh int n, sh int elsize )
{
 sh int p = 0; 
 sh void **temp;
 int i;
 sh int t;
 $ = mpadd(&p,1); // assert consecutive numbering of $ in 0..p-1
 
#ifdef HAVE_IO
 //seq printf("reduce nach %p (%f) von %p size %d elsize %d\n", s, 
 //           *(float *)s, d, n, elsize);
#endif

 temp = (void **)shalloc( p * elsize );  // temporary array
 farm // initialize temp by the first p elements of d:
   if ($<n)
      memcpy( temp+$*elsize, d+$*elsize, elsize );

#ifdef HAVE_IO
#if 0
 seq {
   printf("Operandenvektor: (%d,%d)\n", n, p);
   for (i=0; i<n; i++)
      printf(" %d: loc %d  val %f\n", i, ((pair *)d)[i].loc, ((pair *)d)[i].val );
   for (i=0; i<p; i++)
      printf(" %d: loc %d  val %f\n", i, ((pair *)temp)[i].loc, ((pair *)temp)[i].val );
 }
 barrier;
#endif
#endif

 // p partial reductions being computed:
 farm
   forall( i, p, n, p ) 
     f( temp + $*elsize, temp + $*elsize, d + i*elsize );

 // iterative computation upwards the tree:
 for (t=1; t<n; t = t<<1 )  // sequential loop over tree levels
   farm
     if ($+t < n) 
       f( temp+$*elsize, temp+$*elsize, temp + ($+t)*elsize );
 seq
   memcpy( s, temp, elsize );    // write result to *s
 shallfree();  // release temp
}



async void fabsmax( void *m, void *x, void *y )
{
 if (fabs(((pair*)x)->val) >= fabs(((pair*)y)->val)) {
    ((pair*)m)->val = ((pair *)x)->val;  ((pair*)m)->loc = ((pair*)x)->loc;
 }
 else {
    ((pair*)m)->val = ((pair *)y)->val;  ((pair*)m)->loc = ((pair*)y)->loc;
 }
}


sync void pivotize(sh float **a, sh int n, sh int r)
{
  sh float *hlp;
  sh int p = 0;
  sh pair piv;
  sh pair *col = (pair *)shalloc((n - r)*sizeof(pair));
  int i;
  $ = mpadd(&p, 1);
  for (i = $; i < n - r; i += p) {
     col[i].loc = r + i;
     col[i].val = a[i + r][r];
  }
  reduce(fabsmax, &piv, (void **)col, n - r, sizeof(pair));

#ifdef HAVE_IO
  seq printf("Pivot row is %d, pivot is %f\n", piv.loc, piv.val);
#endif

  farm assert(piv.loc >= r);
  //farm assert( fabs(piv.val - a[piv.loc][r]) < 0.0001 );

  // permute rows:
  hlp = a[r];
  a[r] = a[piv.loc];
  a[piv.loc] = hlp;
}


sync void
getrf(sh float **a, sh int n)
{
  /* applies Gaussian elimination to system a[n][n].
   * Assume that matrix a is regular and diagonally dominant.
   * For a system of n equations in n variables,
   * up to n^2 processors could be used to update in parallel.
   * The solution is returned in the right hand side column a[*][n] 
   * of the modified system tableau.
   */
 int i, j, ij;
 sh int r;
 sh int p = 0;
 sh float factor;
 sh float *rowfactor = (float *)shalloc(n * sizeof(float));
 $ = mpadd(&p, 1);   // renumber
 
 for (r = 0; r < n - 1; r++) {

#ifdef HAVE_IO
    farm if ($==0) print_system(a, n);
#endif

    pivotize(a, n, r);

    factor = 1.0 / a[r][r];

#ifdef HAVE_IO
    seq printf("Stage %d factor %lf p=%d\n", r, factor, p);
#endif

    farm
     forall (i, r + 1, n, p)
       a[i][r] *= factor;

    farm
     forall (i, r + 1, n, p)
       rowfactor[i] = a[i][r];   // copying avoids data dependencies

    farm {
     forall2(i, j, ij, r + 1, n, r + 1, n, p) {
      a[i][j] -= a[r][j] * rowfactor[i];
     }
    }

 }  /* for r */

 return;
}


sh float **s;  /* system tableau */
sh int n = 8;  /* n eq's in n var's */

void main(int ac, char *av[])
{
  pr int starttime, stoptime;

    if (!$) {
        if (ac > 1) {
            n = atoi(av[1]);
        }
    }

    barrier;
    if (n <= 0)
        exit(1);

  start {
    s = build_system( n );

#ifdef HAVE_IO
    farm if ($==0) print_system(s, n);
#endif

    farm starttime = getct();
    getrf( s, n );

#ifdef HAVE_IO
    farm if ($==0) print_system(s, n);
#endif

  }
  stoptime = getct();

#ifdef HAVE_IO
  if ($==0)
    printf("\nTime: %d PRAM CPU cycles\n", stoptime - starttime);
#endif

  barrier;
}
